/* fixowner.c - change the ownership of a file to the owner of the directory
   that contains the file;
   the program must be run either by target user (the directory owner), root
   or nobody; the file to be changed must be initially owned by nobody
   usage:  fixowner [-d] pathname ...
   argument(s) may be either a file or a directory; if it's a directory, all files
   in that directory owned by nobody will be changed, unless -d option is
   specified;
   option: -d - if 'pathname' is a directory, change the ownership of the directory
                itself, instead of all files in it

   13.06.2011 version 1.1: doesn't compile under the recent version of Debian
   because of a few missing #include's - added them
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <pwd.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#include <dirent.h>

#ifdef HAVE_NO_STRLCPY
/* two useful string functions that are defined in some systems (eg. Solaris)
   and aren't in others (eg. Linux) */
   
extern size_t strlcpy(char *dst, const char *src, size_t dstsize);
extern size_t strlcat(char *dst, const char *src, size_t dstsize);
#endif
   
/* #define SYSLOG(fac,...) syslog(fac, __VA_ARGS__); if (isatty(STDERR_FILENO)) { fprintf(stderr,"%s: ",progname); fprintf(stderr, __VA_ARGS__); fprintf(stderr, "\n"); } */
#define SYSLOG(fac,...) { fprintf(stderr,"%s: ",progname); fprintf(stderr, __VA_ARGS__); fprintf(stderr, "\n"); }

/* change the value if your webserver user is other than "nobody" */
#define NOBODY "nobody"

extern struct passwd *get_my_pwent();
extern char *Basename(char *str);

char *progname;
uid_t nobody_uid, my_uid, dir_uid;

char *getpwd(char *dir) {
/* get the canonical absolute pathname for a given directory */
char *oldcwd,*res;

res=NULL;
oldcwd=getcwd(NULL,512);
if (oldcwd!=NULL) {
  if (chdir(dir)==0) {
    res=getcwd(NULL,512);
    chdir(oldcwd);
    free(oldcwd);
  }
}
return res;
}

#define FREE(p,q) if (p!=q) { free(p); }

char *parentdir(char *pathname) {
/* return parent directory of the pathname */
char *c,*p,*q,*res;
struct stat statbuf;
int rc;

if (pathname==NULL) { return NULL; }
#ifdef DEBUG
SYSLOG(LOG_INFO,"obtaining parent directory for %s",pathname);
#endif
rc=stat(pathname,&statbuf);
if (rc==0 && S_ISDIR(statbuf.st_mode)) {   /* is a directory */
  p=getpwd(pathname);
} else {
  p=pathname;
}
c=strrchr(p,'/');
if (c==p) {
  FREE(p,pathname);
  return strdup("/");  // '/' only at the beginning of path -> root dir
}
if (c!=NULL) {
  if (p==pathname) {
    p=strdup(pathname);
    if (p==NULL) { return NULL; }
    p[c-pathname]='\0';
    res=getpwd(p);
    free(p);
  } else {
    *c='\0';
    res=p;
  }
} else {
  FREE(p,pathname);
  res=getcwd(NULL,512);
}
#ifdef DEBUG
SYSLOG(LOG_INFO,"got parent pathname: %s",res==NULL ? "NULL":res);
#endif
return res;
}

void handle_error(char *path) {
if (errno==ENOENT) {    /* not exist */
  SYSLOG(LOG_ERR,"%s does not exist",path);
} else {    /* path access error */
  SYSLOG(LOG_ERR,"access error for %s, error code %d",path,errno);
}
}

uid_t get_dir_uid(char *path, int dir) {
/* if dir is false, and path is a directory, gets its uid (ie, its owner's);
   if dir is false, and path is not a directory, gets uid of directory part of the path;
   if dir is true, always gets uid of directory part of the path */
struct stat statbuf;
int rc; char *p;

rc=lstat(path,&statbuf);
if (rc==0) {    /* exists */
  if (S_ISDIR(statbuf.st_mode) && !dir) {   /* is a directory */
#ifdef DEBUG
SYSLOG(LOG_INFO,"%s is a directory and -d not specified",path);
SYSLOG(LOG_INFO,"determining UID for %s = %ld",path,statbuf.st_uid);
#endif
    return statbuf.st_uid;
  } else {   /* not a directory, or a directory with -d */
    p=parentdir(path);
    if (p==NULL) {
      SYSLOG(LOG_ERR,"can't determine directory pathname for %s",path);
      return -1;
    } else {
#ifdef DEBUG
SYSLOG(LOG_INFO,"determining UID for parent %s",p);
#endif
      rc=lstat(p,&statbuf);
      if (rc==0) {
        free(p);
#ifdef DEBUG
SYSLOG(LOG_INFO,"UID = %ld",statbuf.st_uid);
#endif
        return statbuf.st_uid;
      } else {
        handle_error(p);
        free(p);
        return -1;
      }
    }
  }
} else {
  handle_error(path);
  return -1;
}
}

int fixowner(char *path, int dir_option) {
/* dir_option = 0 -> process all files in directory (first call without -d);
   dir_option = 1 -> skip directories (recursive calls without -d);
   dir_option = 2 -> process directory itself (with -d)

   returns system error code or 250, 251 for special cases
*/
struct stat statbuf;
int rc;

#ifdef DEBUG
SYSLOG(LOG_INFO,"fixing ownership for %s",path);
#endif
rc=lstat(path,&statbuf);
if (rc==0) {    /* exists */
  if (S_ISDIR(statbuf.st_mode) && dir_option<2) {   /* is a directory */
    if (dir_option==1) {
      SYSLOG(LOG_WARNING,"%s is a directory - skipped",path);
      return 250;
    } else {
      /* loop over the directory */
      DIR *dirp; struct dirent *dp; int en,l; char *ptr;
#define MAXBUF 256
      char buf[MAXBUF];
      
      if (strlcpy(buf,path,MAXBUF)>=MAXBUF) {
        SYSLOG(LOG_ERR,"internal error - no memory for directory name");
        return ENOMEM;
      }
      if (buf[strlen(buf)-1]!='/' && strlcat(buf,"/",MAXBUF)>=MAXBUF) {
        SYSLOG(LOG_ERR,"internal error - no memory for directory name");
        return ENOMEM;
      }
      l=MAXBUF-strlen(buf); ptr=buf+strlen(buf);
      dirp=opendir(path);
      if (dirp==NULL) {
        SYSLOG(LOG_ERR,"error reading directory %s, error code %d",path,errno);
        return errno;
      }
      errno=0;
      while ((dp = readdir(dirp)) != NULL) {
        if (strcmp(dp->d_name, ".") == 0 || strcmp(dp->d_name, "..") == 0) { continue; }
        if (strlcpy(ptr,dp->d_name,l)>=l) {
          SYSLOG(LOG_WARNING,"internal error - no memory for filename %s",dp->d_name);
        } else {
          /* call itself recursively */
          (void) fixowner(buf,1);
          errno=0;   // fixowner may fail, so clear errno after it
        }
      }
      en=errno;
      closedir(dirp);
      if (en!=0) {
        SYSLOG(LOG_ERR,"error reading directory %s, error code %d",path,en);
      }
      return en;
    }
  } else if (!S_ISREG(statbuf.st_mode) && !(dir_option==2 && S_ISDIR(statbuf.st_mode))) {   /* not a regular file */
    SYSLOG(LOG_WARNING,"%s not a regular file",path);
    return 250;
  } else {   /* a regular file, or a directory with -d */
    if (statbuf.st_uid==dir_uid) {       /* already correct owner */
#ifdef DEBUG
SYSLOG(LOG_INFO,"%s already has correct owner",path);
#endif
      return 0;
    }
    if (statbuf.st_uid!=nobody_uid) {    /* incorrect ownership */
      SYSLOG(LOG_WARNING,"%s has bad owner: %ld - not changed",path,statbuf.st_uid);
      return 251;
    }
    rc=chown(path,dir_uid,-1);
    if (rc!=0) {
      SYSLOG(LOG_ERR,"can't set owner for %s: error code %d",path,errno);
      return errno;
    }
    return 0;
  }
} else {
  handle_error(path);
  return errno;
}
}


main (int argc, char *argv[]) {

struct passwd *pw;	/* Password file entry for user      */
int dir_option,rc;
char c;

progname=Basename(argv[0]);
/* openlog(progname,LOG_PID,LOG_LOCAL2); */

	/*
	 * Now I have to get the user name.  The name will be gotten 
	 * from the command line if possible.  Otherwise it is figured
	 * out from the environment.
	 */

	pw = get_my_pwent();
	if (!pw) {
           SYSLOG(LOG_ERR,"problem getting my UID");
	   exit(254);
	}
	my_uid = pw->pw_uid;

	pw=getpwnam(NOBODY);
	if (!pw) {
	   SYSLOG(LOG_ERR,"can't determine UID for '%s'",NOBODY)
	   exit(254);
        }
        nobody_uid=pw->pw_uid;

dir_option=0;
while ((c = getopt(argc, argv, "d")) != -1) {
  if (c=='d') dir_option=2;
}

if (optind >= argc) {
  SYSLOG(LOG_ERR,"no filenames specified");
  exit(255);
}
for ( ; optind < argc; optind++) {
#ifdef DEBUG
SYSLOG(LOG_INFO,"Processing: %s",argv[optind]);
#endif
  rc=0;
  dir_uid=get_dir_uid(argv[optind],dir_option);  
  if (dir_uid==-1) {
    rc=errno;
  } else {
    if (my_uid==0 || my_uid==nobody_uid || my_uid==dir_uid) {
      rc=fixowner(argv[optind],dir_option);
    } else {
      SYSLOG(LOG_ERR,"UID of current user (%ld) doesn't match the directory UID for '%s' (%ld)",my_uid,argv[optind],dir_uid);
      rc=EACCES;
    }
  }
}
exit(rc);
}
